"""Services for the job_script_templates resource, including module specific business logic."""
from typing import Any

from buzz import require_condition
from fastapi import status
from sqlalchemy import not_
from sqlalchemy.sql.expression import Select

from jobbergate_api.apps.job_script_templates.models import (
    JobScriptTemplate,
    JobScriptTemplateFile,
    WorkflowFile,
)
from jobbergate_api.apps.services import CrudModel, CrudService, FileService, ServiceError


class JobScriptTemplateService(CrudService):
    """
    Provide a CrudService that overloads the list query builder and locator logic.
    """

    def build_list_query(
        self,
        sort_ascending: bool = True,
        search: str | None = None,
        sort_field: str | None = None,
        include_archived: bool = True,
        include_files: bool = False,
        include_parent: bool = False,
        include_null_identifier: bool = True,
        **additional_filters,
    ) -> Select:
        """List all job_script_templates."""
        query: Select = super().build_list_query(
            sort_ascending=sort_ascending,
            search=search,
            sort_field=sort_field,
            include_archived=include_archived,
            include_files=include_files,
            include_parent=include_parent,
            **additional_filters,
        )
        if not include_null_identifier:
            query = query.where(not_(JobScriptTemplate.identifier.is_(None)))
        return query

    async def create(self, **incoming_data) -> CrudModel:
        """
        Add a new row for the model to the database.
        """
        self.validate_identifier(incoming_data.get("identifier"))
        return await super().create(**incoming_data)

    async def update(self, locator: Any, **incoming_data) -> CrudModel:
        """
        Update a row by locator with supplied data.
        """
        self.validate_identifier(incoming_data.get("identifier"))
        return await super().update(locator, **incoming_data)

    def locate_where_clause(self, id_or_identifier: Any) -> Any:
        """
        Locate an instance using the id or identifier field.
        """
        if isinstance(id_or_identifier, str):
            return JobScriptTemplate.identifier == id_or_identifier
        elif isinstance(id_or_identifier, int):
            return JobScriptTemplate.id == id_or_identifier
        else:
            raise ValueError("id_or_identifier must be a string or integer")

    def validate_identifier(self, identifier: str | None) -> None:
        """
        Validate that the identifier is not an empty string nor composed only by digits.

        Raise a ServiceError with status code 422 if the validation fails.

        Many of the job-script-template endpoints use the id or identifier interchangeably
        as a path parameter. With that, we need to ensure that the identifier is
        not a number, as that would be identified as id.
        """
        if identifier is not None:
            require_condition(
                identifier.strip() and not identifier.isdigit(),
                "Identifier on {} can not be a empty string nor be composed only by digits. Got {}".format(
                    self.name, identifier
                ),
                raise_exc_class=ServiceError,
                raise_kwargs=dict(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY),
            )


class JobScriptTemplateFileService(FileService):
    """
    Provide an empty derived class of FileService.

    Although it doesn't do anything, it fixes errors with mypy:
        error: Value of type variable "FileModel" of "FileService" cannot be "JobScriptTemplateFile"
        error: Value of type variable "FileModel" of "FileService" cannot be "WorkflowFile"
    """


crud_service = JobScriptTemplateService(model_type=JobScriptTemplate)
template_file_service = JobScriptTemplateFileService(model_type=JobScriptTemplateFile)
workflow_file_service = JobScriptTemplateFileService(model_type=WorkflowFile)



"""Router for the Job Script Template resource."""
from typing import cast

from fastapi import APIRouter, BackgroundTasks, Depends, File, HTTPException, Path, Query
from fastapi import Response as FastAPIResponse
from fastapi import UploadFile, status
from fastapi.responses import StreamingResponse
from fastapi_pagination import Page
from loguru import logger

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.dependencies import SecureService, secure_services
from jobbergate_api.apps.garbage_collector import garbage_collect
from jobbergate_api.apps.job_script_templates.models import JobScriptTemplate
from jobbergate_api.apps.job_script_templates.services import crud_service as template_crud_service
from jobbergate_api.apps.job_script_templates.services import template_file_service
from jobbergate_api.apps.job_scripts.models import JobScriptFile
from jobbergate_api.apps.job_scripts.schemas import (
    JobScriptCreateRequest,
    JobScriptDetailedView,
    JobScriptListView,
    JobScriptUpdateRequest,
    RenderFromTemplateRequest,
)
from jobbergate_api.apps.job_scripts.services import crud_service, file_service
from jobbergate_api.apps.job_scripts.tools import inject_sbatch_params
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.apps.schemas import ListParams
from jobbergate_api.storage import secure_session

router = APIRouter(prefix="/job-scripts", tags=["Job Scripts"])


@router.delete(
    "/clean-unused-entries",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint to automatically clean unused job scripts depending on a threshold",
    tags=["Garbage collector"],
    dependencies=[Depends(secure_services(Permissions.JOB_SCRIPTS_EDIT))],
)
def job_script_auto_clean_unused_entries(background_tasks: BackgroundTasks):
    """Automatically clean unused job scripts depending on a threshold."""
    logger.info("Starting automatically cleanup for unused job scripts")
    background_tasks.add_task(crud_service.auto_clean_unused_job_scripts)
    return {"description": "Automatically cleanup started"}


@router.post(
    "",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptDetailedView,
    description="Endpoint for creating a stand alone job script. Use file upload to add files.",
)
async def job_script_create(
    create_request: JobScriptCreateRequest,
    secure_services: SecureService = Depends(secure_services(Permissions.JOB_SCRIPTS_EDIT)),
):
    """Create a stand alone job script."""
    logger.info(f"Creating a new job script with {create_request=}")

    if secure_services.identity_payload.email is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="The token payload does not contain an email",
        )

    return await crud_service.create(
        owner_email=secure_services.identity_payload.email,
        **create_request.dict(exclude_unset=True),
    )


@router.post(
    "/render-from-template/{id_or_identifier}",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptDetailedView,
    description="Endpoint for job script creation",
)
async def job_script_create_from_template(
    create_request: JobScriptCreateRequest,
    render_request: RenderFromTemplateRequest,
    id_or_identifier: int | str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SCRIPTS_EDIT, ensure_email=True)
    ),
):
    """Create a new job script from a job script template."""
    logger.info(f"Creating a new job script from {id_or_identifier=} with {create_request=}")

    base_template = cast(
        JobScriptTemplate, await template_crud_service.get(id_or_identifier, include_files=True)
    )

    required_map = render_request.template_output_name_mapping
    entrypoint_key = list(required_map.values())[0]

    required_keys = set(required_map.keys())
    provided_keys = set(f.filename for f in base_template.template_files)
    missing_keys = required_keys - provided_keys
    if len(missing_keys) > 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=(f"The template files are missing required files: {missing_keys} for {id_or_identifier}"),
        )

    mapped_template_files = {
        new_filename: file
        for file in base_template.template_files
        if (new_filename := required_map.get(file.filename, None))
    }

    job_script = await crud_service.create(
        owner_email=secure_services.identity_payload.email,
        parent_template_id=base_template.id,
        **create_request.dict(exclude_unset=True),
    )

    for new_filename, template_file in mapped_template_files.items():
        file_content = await template_file_service.render(
            template_file,
            render_request.param_dict,
        )

        if new_filename == entrypoint_key:
            file_type = FileType.ENTRYPOINT
            if render_request.sbatch_params:
                file_content = inject_sbatch_params(file_content, render_request.sbatch_params)
        else:
            file_type = FileType.SUPPORT

        if template_file.file_type != file_type:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"The file type {template_file} does not match the expected file type {file_type}",
            )

        await file_service.upsert(
            parent_id=job_script.id,
            filename=new_filename,
            upload_content=file_content,
            file_type=file_type,
        )

    return await crud_service.get(job_script.id, include_files=True)


@router.get(
    "/{id}",
    description="Endpoint to return a job script by its id",
    response_model=JobScriptDetailedView,
    dependencies=[Depends(secure_services(Permissions.JOB_SCRIPTS_VIEW))],
)
async def job_script_get(id: int = Path()):
    """Get a job script by id."""
    logger.info(f"Getting job script {id=}")
    return await crud_service.get(id, include_files=True)


@router.get(
    "",
    description="Endpoint to return a list of job scripts",
    response_model=Page[JobScriptListView],
)
async def job_script_get_list(
    list_params: ListParams = Depends(),
    from_job_script_template_id: int
    | None = Query(
        None,
        description="Filter job-scripts by the job-script-template-id they were created from.",
    ),
    secure_services: SecureService = Depends(secure_services(Permissions.JOB_SCRIPTS_VIEW)),
):
    """Get a list of job scripts."""
    logger.debug("Preparing to list job scripts")

    list_kwargs = list_params.dict(exclude_unset=True, exclude={"user_only"})

    if from_job_script_template_id is not None:
        list_kwargs["parent_template_id"] = from_job_script_template_id
    if list_params.user_only:
        list_kwargs["owner_email"] = secure_services.identity_payload.email

    return await crud_service.paginated_list(**list_kwargs)


@router.put(
    "/{id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job script by id",
    response_model=JobScriptListView,
)
async def job_script_update(
    update_params: JobScriptUpdateRequest,
    id: int = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SCRIPTS_EDIT, ensure_email=True)
    ),
):
    """Update a job script template by id or identifier."""
    logger.info(f"Updating job script {id=} with {update_params=}")
    await crud_service.get(id, ensure_attributes={"owner_email": secure_services.identity_payload.email})
    return await crud_service.update(id, **update_params.dict(exclude_unset=True))


@router.delete(
    "/{id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete a job script by id",
)
async def job_script_delete(
    id: int = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SCRIPTS_EDIT, ensure_email=True)
    ),
):
    """Delete a job script template by id or identifier."""
    logger.info(f"Deleting job script {id=}")
    await crud_service.get(id, ensure_attributes={"owner_email": secure_services.identity_payload.email})
    await crud_service.delete(id)
    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{id}/upload/{file_name:path}",
    description="Endpoint to get a file from a job script",
    dependencies=[
        Depends(secure_services(Permissions.JOB_SCRIPTS_VIEW)),
    ],
)
async def job_script_get_file(id: int = Path(...), file_name: str = Path(...)):
    """
    Get a job script file.

    Note:
        See https://fastapi.tiangolo.com/advanced/custom-response/#streamingresponse
    """
    job_script = await crud_service.get(id)
    job_script_file = await file_service.get(job_script.id, file_name)
    return StreamingResponse(
        content=await file_service.stream_file_content(job_script_file),
        media_type="text/plain",
        headers={"filename": job_script_file.filename},
    )


@router.put(
    "/{id}/upload/{file_type}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to upload a file to a job script file",
)
async def job_script_upload_file(
    id: int = Path(...),
    file_type: FileType = Path(...),
    upload_file: UploadFile = File(..., description="File to upload"),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SCRIPTS_EDIT, ensure_email=True)
    ),
):
    """Upload a file to a job script."""
    if upload_file.filename is None:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="The upload file has no filename",
        )

    logger.debug(f"Uploading file {upload_file.filename} to job script {id=}")

    job_script = await crud_service.get(
        id, ensure_attributes={"owner_email": secure_services.identity_payload.email}
    )

    await file_service.upsert(
        parent_id=job_script.id,
        filename=upload_file.filename,
        upload_content=upload_file,
        file_type=file_type,
    )


@router.delete(
    "/{id}/upload/{file_name}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to delete a file from a job script",
)
async def job_script_delete_file(
    id: int = Path(...),
    file_name: str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SCRIPTS_EDIT, ensure_email=True)
    ),
):
    """Delete a file from a job script template by id or identifier."""
    job_script = await crud_service.get(
        id, ensure_attributes={"owner_email": secure_services.identity_payload.email}
    )
    job_script_file = await file_service.get(job_script.id, file_name)
    await file_service.delete(job_script_file)


@router.delete(
    "/upload/garbage-collector",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint to delete all unused files from the job script file storage",
    tags=["Garbage collector"],
)
def job_script_garbage_collector(
    background_tasks: BackgroundTasks,
    secure_services: SecureService = Depends(secure_session(Permissions.JOB_SCRIPTS_EDIT)),
):
    """Delete all unused files from job scripts on the file storage."""
    logger.info("Starting garbage collection from jobbergate file storage")
    background_tasks.add_task(
        garbage_collect,
        secure_services.session,
        secure_services.bucket,
        [JobScriptFile],
        background_tasks,
    )
    return {"description": "Garbage collection started"}



"""
Provide schemas for the job script templates component.
"""

import json
from datetime import datetime
from typing import Any, Optional

import pydantic
from pydantic import BaseModel

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.schemas import TableResource
from jobbergate_api.meta_mapper import MetaField, MetaMapper

job_template_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    parent_id=MetaField(
        description="The unique database identifier for the parent of this instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2023-08-18T13:55:37.172285",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2023-08-18T13:55:37.172285",
    ),
    name=MetaField(
        description="The unique name of the instance",
        example="test-job-script-88",
    ),
    identifier=MetaField(
        description="A human-friendly label used for lookup on frequently accessed applications",
        example="App88",
    ),
    description=MetaField(
        description="A text field providing a human-friendly description of the job_script",
        example="This job_scripts runs an Foo job using the bar variant",
    ),
    owner_email=MetaField(
        description="The email of the owner/creator of the instance",
        example="tucker@omnivector.solutions",
    ),
    template_vars=MetaField(
        description="The template variables of the job script template",
        example={"param1": 7, "param2": 13},
    ),
    runtime_config=MetaField(
        description="The runtime configuration of the workflow",
        example={"param1": 7, "param2": 13},
    ),
    filename=MetaField(
        description="The name of the file",
        example="job-script.py",
    ),
    file_type=MetaField(
        description="The type of the file",
        example=FileType.ENTRYPOINT.value,
    ),
    template_files=MetaField(
        description="The template files attached to a job script template",
    ),
    workflow_file=MetaField(
        description="The workflow file attached to a job script template",
    ),
    is_archived=MetaField(
        description="Indicates if the job script template has been archived.",
        example=False,
    ),
)


class RunTimeConfig(BaseModel):
    """
    Schema for the runtime config of a job template.

    Notice this includes user supplied variables, so it has no predefined field.
    It also loads the contend directly from the json at the request payload.
    """

    @classmethod
    def __get_validators__(cls):
        """
        Get the validators.
        """
        yield cls.validate_to_json

    @classmethod
    def validate_to_json(cls, value):
        """
        Validate the produced json.
        """
        if isinstance(value, str):
            return cls(**json.loads(value))
        return value

    class Config:
        extra = pydantic.Extra.allow
        schema_extra = job_template_meta_mapper


class JobTemplateCreateRequest(BaseModel):
    """Schema for the request to create a job template."""

    name: str
    identifier: str | None
    description: str | None
    template_vars: dict[str, Any] | None

    class Config:
        schema_extra = job_template_meta_mapper


class JobTemplateUpdateRequest(BaseModel):
    """Schema for the request to update a job template."""

    name: str | None
    identifier: str | None
    description: str | None
    template_vars: dict[str, Any] | None
    is_archived: bool | None

    class Config:
        schema_extra = job_template_meta_mapper


class TemplateFileDetailedView(BaseModel):
    """Schema for the response to get a template file."""

    parent_id: int
    filename: str
    file_type: FileType
    created_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True
        schema_extra = job_template_meta_mapper


class WorkflowFileDetailedView(BaseModel):
    """Schema for the response to get a workflow file."""

    parent_id: int
    filename: str
    runtime_config: Optional[dict[str, Any]] = {}
    created_at: Optional[datetime]
    updated_at: Optional[datetime]

    class Config:
        orm_mode = True
        schema_extra = job_template_meta_mapper


class JobTemplateListView(TableResource):
    """Schema for the response to get a list of entries."""

    identifier: Optional[str]

    class Config:
        schema_extra = job_template_meta_mapper


class JobTemplateDetailedView(JobTemplateListView):
    """
    Schema for the request to an entry.

    Notice the files default to None, as they are not always requested, to differentiate between
    an empty list when they are requested, but no file is found.
    """

    template_vars: dict[str, Any] | None

    template_files: list[TemplateFileDetailedView] | None
    workflow_files: list[WorkflowFileDetailedView] | None



"""
Main file to startup the fastapi server.
"""
import sys
import typing
from contextlib import asynccontextmanager

import asyncpg
import sentry_sdk
from fastapi import FastAPI, HTTPException, Request, Response, status
from fastapi.exceptions import RequestValidationError
from fastapi_pagination import add_pagination
from loguru import logger
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from starlette.middleware.cors import CORSMiddleware

from jobbergate_api import __version__
from jobbergate_api.apps.job_script_templates.routers import router as job_script_templates_router
from jobbergate_api.apps.job_scripts.routers import router as job_scripts_router
from jobbergate_api.apps.job_submissions.routers import router as job_submissions_router
from jobbergate_api.config import settings
from jobbergate_api.storage import engine_factory, handle_fk_error

subapp = FastAPI(
    title="Jobbergate-API",
    version=__version__,
    contact={
        "name": "Omnivector Solutions",
        "url": "https://www.omnivector.solutions/",
        "email": "info@omnivector.solutions",
    },
    license_info={
        "name": "MIT License",
        "url": "https://github.com/omnivector-solutions/jobbergate/blob/main/LICENSE",
    },
)

subapp.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

if settings.SENTRY_DSN and settings.DEPLOY_ENV.lower() != "test":
    logger.info("Initializing Sentry")
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        sample_rate=typing.cast(float, settings.SENTRY_SAMPLE_RATE),  # The cast silences mypy
        environment=settings.DEPLOY_ENV,
    )
    subapp.add_middleware(SentryAsgiMiddleware)
else:
    logger.info("Skipping Sentry")

subapp.include_router(job_script_templates_router)
subapp.include_router(job_scripts_router)
subapp.include_router(job_submissions_router)
subapp.exception_handler(asyncpg.exceptions.ForeignKeyViolationError)(handle_fk_error)

add_pagination(subapp)


@subapp.get(
    "/health",
    status_code=status.HTTP_204_NO_CONTENT,
    responses={204: {"description": "API is healthy"}},
)
async def health_check():
    """
    Provide a health-check endpoint for the app.
    """
    logger.debug("CHECKING HEALTH")
    return Response(status_code=status.HTTP_204_NO_CONTENT)


@asynccontextmanager
async def lifespan(_: FastAPI):
    """
    Provide a lifespan context for the app.

    Will set up logging and cleanup database engines when the app is shut down.

    This is the preferred method of handling lifespan events in FastAPI.
    For mor details, see: https://fastapi.tiangolo.com/advanced/events/
    """
    logger.remove()
    logger.add(sys.stderr, level=settings.LOG_LEVEL)
    logger.info(f"Logging configured 📝 Level: {settings.LOG_LEVEL}")

    yield

    # Skip cleanup if in test mode so that model instances can be used after processing http requests.
    if settings.DEPLOY_ENV.lower() != "test":
        await engine_factory.cleanup()


app = FastAPI(lifespan=lifespan)
app.mount("/jobbergate", subapp)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, err: RequestValidationError):
    """
    Handle exceptions from pydantic validators.
    """
    raise HTTPException(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        detail=f"Validation error for request to {request.url} with data: {request.json()}: {err}",
    )



"""
Provide configuration settings for the app.

Pull settings from environment variables or a .env file if available.
"""
from enum import Enum
from typing import Optional

from buzz import require_condition
from loguru import logger
from pydantic import BaseSettings, Field, HttpUrl, root_validator


class LogLevelEnum(str, Enum):
    """
    Provide an enumeration class describing the available log levels.
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def check_none_or_all_keys_exist(input_dict: dict, target_keys: set) -> bool:
    """
    Verify if none or all of the target keys exist in the input dictionary.
    """
    all_exist = all(k in input_dict for k in target_keys)
    none_exist = all(k not in input_dict for k in target_keys)
    return all_exist or none_exist


class Settings(BaseSettings):
    """
    Provide a pydantic ``BaseSettings`` model for the application settings.
    """

    DEPLOY_ENV: str = "LOCAL"

    LOG_LEVEL: LogLevelEnum = LogLevelEnum.DEBUG

    # Database settings  # Default to values from docker-compose.yml
    DATABASE_HOST: str = "localhost"
    DATABASE_USER: str = "local-user"
    DATABASE_PSWD: str = "local-pswd"
    DATABASE_NAME: str = "local-db"
    DATABASE_PORT: int = 5432

    # Test database settings
    TEST_DATABASE_HOST: str = "localhost"
    TEST_DATABASE_USER: str = "test-user"
    TEST_DATABASE_PSWD: str = "test-pswd"
    TEST_DATABASE_NAME: str = "test-db"
    TEST_DATABASE_PORT: int = 5433

    # S3 configuration
    S3_BUCKET_NAME: str = Field("jobbergate-staging-eu-north-1-resources")
    S3_ENDPOINT_URL: Optional[str]

    # Test S3 configuration
    TEST_S3_BUCKET_NAME: str = Field("test-jobbergate-resources")
    TEST_S3_ENDPOINT_URL: str = Field("http://localhost:9000")

    # Security Settings. For details, see https://github.com/omnivector-solutions/armasec
    ARMASEC_DOMAIN: str
    ARMASEC_USE_HTTPS: bool = Field(True)
    ARMASEC_AUDIENCE: Optional[HttpUrl]
    ARMASEC_DEBUG: bool = Field(False)
    ARMASEC_ADMIN_DOMAIN: Optional[str]
    ARMASEC_ADMIN_AUDIENCE: Optional[HttpUrl]
    ARMASEC_ADMIN_MATCH_KEY: Optional[str]
    ARMASEC_ADMIN_MATCH_VALUE: Optional[str]

    # Key to custom claims packaged with Auth0 tokens
    IDENTITY_CLAIMS_KEY: str = "https://omnivector.solutions"

    # Sentry configuration
    SENTRY_DSN: Optional[HttpUrl]
    SENTRY_SAMPLE_RATE: float = Field(1.0, gt=0.0, le=1.0)

    # Maximum number of bytes allowed for file uploads
    MAX_UPLOAD_FILE_SIZE: int = 100 * 1024 * 1024  # 100 MB

    # Sendgrid configuration for email notification
    SENDGRID_FROM_EMAIL: Optional[str]
    SENDGRID_API_KEY: Optional[str]

    # Enable multi-tenancy so that the database is determined by the client_id in the auth token
    MULTI_TENANCY_ENABLED: bool = Field(False)

    # Automatically clean up unused job scripts
    AUTO_CLEAN_JOB_SCRIPTS_DAYS_TO_ARCHIVE: int | None
    AUTO_CLEAN_JOB_SCRIPTS_DAYS_TO_DELETE: int | None

    @root_validator(pre=True)
    def remove_blank_env(cls, values):
        """
        Remove any settings from the environment that are blank strings.

        This allows the defaults to be set if ``docker-compose`` defaults a missing
        environment variable to a blank string.
        """
        clean_values = dict()
        for key, value in values.items():
            if isinstance(value, str):
                if value.strip():
                    clean_values[key] = value
                else:
                    logger.warning(f"The setting {key} is a blank string and was not considered.")
            else:
                clean_values[key] = value

        require_condition(
            check_none_or_all_keys_exist(
                clean_values,
                {"SENDGRID_FROM_EMAIL", "SENDGRID_API_KEY"},
            ),
            "Either none or all SendGrind parameters are expected.",
            RuntimeError,
        )

        return clean_values

    class Config:
        env_file = ".env"


settings = Settings()



"""
Router dependencies shared for multiple resources.

Note:
    The dependencies can be reused multiple times, since FastAPI caches the results.
"""
from contextlib import asynccontextmanager, contextmanager
from dataclasses import dataclass
from typing import AsyncIterator, Iterator

from aioboto3.session import Session
from fastapi import Depends

from jobbergate_api.apps.services import BucketBoundService, DatabaseBoundService
from jobbergate_api.config import settings
from jobbergate_api.safe_types import Bucket
from jobbergate_api.security import PermissionMode
from jobbergate_api.storage import AsyncSession, SecureSession, secure_session

session = Session()


@asynccontextmanager
async def s3_bucket(bucket_name: str, s3_url: str | None) -> AsyncIterator[Bucket]:
    """Create a bucket using a context manager."""
    async with session.resource("s3", endpoint_url=s3_url) as s3:
        bucket = await s3.Bucket(bucket_name)
        yield bucket


def get_bucket_name(override_bucket_name: str | None = None) -> str:
    """
    Get the bucket name based on the environment.

    The name can be overridden when multi tenancy is enabled by passing a bucket name.
    """
    if settings.DEPLOY_ENV.lower() == "test":
        return settings.TEST_S3_BUCKET_NAME
    if override_bucket_name and settings.MULTI_TENANCY_ENABLED:
        return override_bucket_name
    return settings.S3_BUCKET_NAME


def get_bucket_url() -> str | None:
    """Get the bucket url based on the environment."""
    if settings.DEPLOY_ENV.lower() == "test":
        return settings.TEST_S3_ENDPOINT_URL
    return settings.S3_ENDPOINT_URL


@contextmanager
def bind_session(session: AsyncSession) -> Iterator[None]:
    """Bind the session to all CRUD services."""
    try:
        [service.bind_session(session) for service in DatabaseBoundService.database_services]
        yield
    finally:
        [service.unbind_session() for service in DatabaseBoundService.database_services]


@contextmanager
def bind_bucket(bucket: Bucket) -> Iterator[None]:
    """Bind the bucket to all file services."""
    try:
        [service.bind_bucket(bucket) for service in BucketBoundService.bucket_services]
        yield
    finally:
        [service.unbind_bucket() for service in BucketBoundService.bucket_services]


@dataclass
class SecureService(SecureSession):
    """Dataclass to hold the secure session and the bucket."""

    bucket: Bucket


def secure_services(
    *scopes: str,
    permission_mode: PermissionMode = PermissionMode.ALL,
    ensure_email: bool = False,
    ensure_client_id: bool = False,
):
    """
    Dependency to bind database services to a secure session.
    """

    async def dependency(
        secure_session: SecureSession = Depends(
            secure_session(
                *scopes,
                permission_mode=permission_mode,
                ensure_email=ensure_email,
                ensure_organization=settings.MULTI_TENANCY_ENABLED is True,
                ensure_client_id=ensure_client_id,
            )
        ),
    ) -> AsyncIterator[SecureService]:
        """
        Bind each service to the secure session and then return the session.
        """
        bucket_name = get_bucket_name(secure_session.identity_payload.organization_id)
        s3_url = get_bucket_url()

        async with s3_bucket(bucket_name, s3_url) as bucket:
            with bind_session(secure_session.session), bind_bucket(bucket):
                yield SecureService(
                    identity_payload=secure_session.identity_payload,
                    session=secure_session.session,
                    bucket=bucket,
                )

    return dependency



"""
JobSubmission resource schema.
"""
from pathlib import Path
from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, Extra, Field

from jobbergate_api.apps.job_scripts.schemas import JobScriptListView
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus
from jobbergate_api.apps.schemas import TableResource
from jobbergate_api.meta_mapper import MetaField, MetaMapper

job_submission_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2023-08-18T13:55:37.172285",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2023-08-18T13:55:37.172285",
    ),
    name=MetaField(
        description="The unique name of the job submission",
        example="test-job-submission-77",
    ),
    description=MetaField(
        description="A text field providing a human-friendly description of the job_submission",
        example="Submission for the Foo job on sample 99 using the bar variant",
    ),
    owner_email=MetaField(
        description="The email of the owner/creator of the instance",
        example="tucker@omnivector.solutions",
    ),
    job_script_id=MetaField(
        description="The foreign-key to the job_script from which this instance was created",
        example=71,
    ),
    execution_directory=MetaField(
        description="The directory on the cluster where the job should be executed",
        example="/home/someuser/job-data/test-77",
    ),
    slurm_job_id=MetaField(
        description="The id for the slurm job executing this job_submission",
        example="1883",
    ),
    client_id=MetaField(
        description="The client_id of the cluster where this job submission should execute",
        example="mega-cluster-1",
    ),
    status=MetaField(
        description=f"The status of the job submission. Must be one of {JobSubmissionStatus.pretty_list()}",
        example=JobSubmissionStatus.CREATED,
    ),
    report_message=MetaField(
        description="The report message received from cluster-agent when a job submission is rejected",
        example="Unrecognized SBATCH arguments",
    ),
    execution_parameters=MetaField(
        description=(
            "The parameters to be passed to the job submission. "
            "See more details at: https://slurm.schedmd.com/rest_api.html"
        ),
        example={"name": "job-submission-name", "comment": "I am a comment"},
    ),
)


class JobProperties(BaseModel, extra=Extra.forbid):
    """
    Specialized model for job properties.

    See more details at: https://slurm.schedmd.com/rest_api.html
    """

    account: Optional[str] = Field(
        description="Charge resources used by this job to specified account.",
    )
    account_gather_frequency: Optional[str] = Field(
        description="Define the job accounting and profiling sampling intervals.",
    )
    argv: Optional[List[str]] = Field(description="Arguments to the script.")
    array: Optional[str] = Field(
        description=(
            "Submit a job array, multiple jobs to be executed with identical parameters. "
            "The indexes specification identifies what array index values should be used."
        )
    )
    batch_features: Optional[str] = Field(description="features required for batch script's node")
    begin_time: Optional[str] = Field(
        description=(
            "Submit the batch script to the Slurm controller immediately, like normal, "
            "but tell the controller to defer the allocation of the job until the specified time."
        )
    )
    burst_buffer: Optional[str] = Field(description="Burst buffer specification.")
    cluster_constraints: Optional[str] = Field(
        description=(
            "Specifies features that a federated cluster must have to have a sibling job submitted to it."
        )
    )
    comment: Optional[str] = Field(description="An arbitrary comment.")
    constraints: Optional[str] = Field(description="node features required by job.")
    core_specification: Optional[int] = Field(
        description=(
            "Count of specialized threads per node reserved by the job for system "
            "operations and not used by the application."
        )
    )
    cores_per_socket: Optional[int] = Field(
        description=(
            "Restrict node selection to nodes with at least the specified number of cores per socket."
        )
    )
    cpu_binding: Optional[str] = Field(description="Cpu binding")
    cpu_binding_hint: Optional[str] = Field(description="Cpu binding hint")
    cpu_frequency: Optional[str] = Field(
        description=(
            "Request that job steps initiated by srun commands inside this sbatch "
            "script be run at some requested frequency if possible, on the CPUs "
            "selected for the step on the compute node(s)."
        ),
    )
    cpus_per_gpu: Optional[str] = Field(description="Number of CPUs requested per allocated GPU.")
    cpus_per_task: Optional[int] = Field(
        description=(
            "Advise the Slurm controller that ensuing job steps will require "
            "ncpus number of processors per task."
        ),
    )
    current_working_directory: Optional[str] = Field(
        description="Instruct Slurm to connect the batch script's standard output directly to the file name."
    )
    deadline: Optional[str] = Field(
        description=(
            "Remove the job if no ending is possible before this deadline (start > (deadline - time[-min]))."
        )
    )
    delay_boot: Optional[int] = Field(
        description=(
            "Do not reboot nodes in order to satisfied this job's feature specification if "
            "the job has been eligible to run for less than this time period."
        )
    )
    dependency: Optional[str] = Field(
        description=(
            "Defer the start of this job until the specified dependencies have been satisfied completed."
        )
    )
    distribution: Optional[str] = Field(
        description="Specify alternate distribution methods for remote processes."
    )
    environment: Optional[Dict[str, str]] = Field(description="Dictionary of environment entries.")
    exclusive: Optional[Literal["user", "mcs", "exclusive", "oversubscribe"]] = Field(
        description=(
            "The job allocation can share nodes just other users with the "
            "'user' option or with the 'mcs' option)."
        )
    )
    get_user_environment: int = Field(
        default=1,
        description="Load new login environment for user on job node.",
        ge=0,
        le=1,
    )
    gres: Optional[str] = Field(
        description="Specifies a comma delimited list of generic consumable resources."
    )
    gres_flags: Optional[Literal["disable-binding", "enforce-binding"]] = Field(
        description="Specify generic resource task binding options."
    )
    gpu_binding: Optional[str] = Field(description="Requested binding of tasks to GPU.")
    gpu_frequency: Optional[str] = Field(description="Requested GPU frequency.")
    gpus: Optional[str] = Field(description="GPUs per job.")
    gpus_per_node: Optional[str] = Field(description="GPUs per node.")
    gpus_per_socket: Optional[str] = Field(description="GPUs per socket.")
    gpus_per_task: Optional[str] = Field(description="GPUs per task.")
    hold: Optional[bool] = Field(
        description="Specify the job is to be submitted in a held state (priority of zero)."
    )
    kill_on_invalid_dependency: Optional[bool] = Field(
        description="If a job has an invalid dependency, then Slurm is to terminate it."
    )
    licenses: Optional[str] = Field(
        description=(
            "Specification of licenses (or other resources available on all nodes of the cluster) "
            "which must be allocated to this job."
        )
    )
    mail_type: Optional[str] = Field(description="Notify user by email when certain event types occur.")
    mail_user: Optional[str] = Field(
        description="User to receive email notification of state changes as defined by mail_type."
    )
    mcs_label: Optional[str] = Field(description="This parameter is a group among the groups of the user.")
    memory_binding: Optional[str] = Field(description="Bind tasks to memory.")
    memory_per_cpu: Optional[int] = Field(description="Minimum real memory per cpu (MB).")
    memory_per_gpu: Optional[int] = Field(description="Minimum memory required per allocated GPU.")
    memory_per_node: Optional[str] = Field(description="Minimum real memory per node (MB).")
    minimum_cpus_per_node: Optional[int] = Field(description="Minimum number of CPUs per node.")
    minimum_nodes: Optional[bool] = Field(
        description="If a range of node counts is given, prefer the smaller count."
    )
    name: Optional[str] = Field(description="Specify a name for the job allocation.")
    nice: Optional[str] = Field(description="Run the job with an adjusted scheduling priority within Slurm.")
    no_kill: Optional[bool] = Field(
        description="Do not automatically terminate a job if one of the nodes it has been allocated fails."
    )
    nodes: Optional[str] = Field(
        description="Request that a minimum of nodes nodes and a maximum node count.",
    )
    open_mode: Optional[Literal["append", "truncate"]] = Field(
        description="Open the output and error files using append or truncate mode as specified."
    )
    partition: Optional[str] = Field(description="Request a specific partition for the resource allocation.")
    priority: Optional[str] = Field(description="Request a specific job priority.")
    qos: Optional[str] = Field(description="Request a quality of service for the job.")
    requeue: Optional[bool] = Field(
        description="Specifies that the batch job should eligible to being requeue."
    )
    reservation: Optional[str] = Field(
        description="Allocate resources for the job from the named reservation."
    )
    signal: Optional[str] = Field(
        description="When a job is within sig_time seconds of its end time, send it the signal sig_num.",
    )
    sockets_per_node: Optional[int] = Field(
        description="Restrict node selection to nodes with at least the specified number of sockets."
    )
    spread_job: Optional[bool] = Field(
        description=(
            "Spread the job allocation over as many nodes as possible and attempt "
            "to evenly distribute tasks across the allocated nodes."
        )
    )
    standard_error: Optional[str] = Field(
        description="Instruct Slurm to connect the batch script's standard error directly to the file name."
    )
    standard_input: Optional[str] = Field(
        description=(
            "Instruct Slurm to connect the batch script's standard input directly to the file name specified."
        )
    )
    standard_output: Optional[str] = Field(
        description="Instruct Slurm to connect the batch script's standard output directly to the file name."
    )
    tasks: Optional[int] = Field(
        description=(
            "Advises the Slurm controller that job steps run within the allocation "
            "will launch a maximum of number tasks and to provide for sufficient resources."
        )
    )
    tasks_per_core: Optional[int] = Field(description="Request the maximum ntasks be invoked on each core.")
    tasks_per_node: Optional[int] = Field(description="Request the maximum ntasks be invoked on each node.")
    tasks_per_socket: Optional[int] = Field(
        description="Request the maximum ntasks be invoked on each socket."
    )
    thread_specification: Optional[int] = Field(
        description=(
            "Count of specialized threads per node reserved by the job for system "
            "operations and not used by the application."
        )
    )
    threads_per_core: Optional[int] = Field(
        description="Restrict node selection to nodes with at least the specified number of threads per core."
    )
    time_limit: Optional[str] = Field(description="Step time limit.")
    time_minimum: Optional[int] = Field(description="Minimum run time in minutes.")
    wait_all_nodes: Optional[int] = Field(
        description="Do not begin execution until all nodes are ready for use.", ge=0, le=1
    )
    wckey: Optional[str] = Field(description="Specify wckey to be used with job.")


class JobSubmissionCreateRequest(BaseModel):
    """
    Request model for creating JobSubmission instances.
    """

    name: str
    description: Optional[str]
    job_script_id: int
    execution_directory: Optional[str]
    client_id: Optional[str]
    execution_parameters: JobProperties = Field(default_factory=JobProperties)

    class Config:
        schema_extra = job_submission_meta_mapper


class JobSubmissionUpdateRequest(BaseModel):
    """
    Request model for updating JobSubmission instances.
    """

    name: Optional[str]
    description: Optional[str]
    execution_directory: Optional[str]
    status: Optional[JobSubmissionStatus]

    class Config:
        schema_extra = job_submission_meta_mapper


class JobSubmissionAgentUpdateRequest(BaseModel):
    """Request model for updating JobSubmission instances."""

    status: JobSubmissionStatus
    slurm_job_id: Optional[int]
    report_message: Optional[str]

    class Config:
        schema_extra = job_submission_meta_mapper


class JobSubmissionListView(TableResource):
    """
    Partial model to match the database for the JobSubmission resource.
    """

    job_script_id: Optional[int]
    slurm_job_id: Optional[int]
    client_id: str
    status: JobSubmissionStatus

    job_script: Optional[JobScriptListView]

    class Config:
        schema_extra = job_submission_meta_mapper


class JobSubmissionDetailedView(JobSubmissionListView):
    """
    Complete model to match the database for the JobSubmission resource.
    """

    execution_directory: Optional[Path]
    report_message: Optional[str]
    execution_parameters: Optional[JobProperties]


class PendingJobSubmission(BaseModel):
    """
    Specialized model for the cluster-agent to pull pending job_submissions.

    Model also includes data from its job_script and application sources.
    """

    id: int
    name: str
    owner_email: str
    execution_directory: Optional[Path]
    execution_parameters: dict = Field(default_factory=dict)
    job_script: Optional[JobScriptListView]

    class Config:
        orm_mode = True
        extra = Extra.ignore
        schema_extra = job_submission_meta_mapper


class ActiveJobSubmission(BaseModel):
    """
    Specialized model for the cluster-agent to pull an active job_submission.
    """

    id: int
    name: str
    slurm_job_id: int

    class Config:
        orm_mode = True
        extra = Extra.ignore
